#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#ifdef __unix__
 #include <unistd.h>
#endif
#ifdef __MSDOS__
 #include <io.h>
#endif
#include <errno.h>
#include "mlist.h"

struct cfg **list_cfg;
int lists = 0;
char *prgname;
char bbs[7], sysop[7];

static int read_cfg(void)
{
	FILE *fp;
	char fname[20], line[128], tmp_flags[4], tmpadr[49];
	int i, lineno = 0, rl = 0;

	build_path(fname, prgname, "CFG");

	if ((fp = fopen(fname, "r")) == NULL) {
		print_error("Error opening input file %s: %s\n", fname, strerror(errno));
		return (-1);
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		lineno++;

		if (line[0] == '#' || line[0] == '\n')
			continue;

		switch (++rl) {
		case 1:
			sscanf(line, "%6s", bbs);
			strupr(bbs);
			continue;
		case 2:
			sscanf(line, "%6s", sysop);
			strupr(sysop);
			continue;
		case 3:
			sscanf(line, "%128s", mailin_file);
			continue;
		}

		list_cfg = (struct cfg **) do_realloc(list_cfg, sizeof(struct cfg *) * (lists + 1));
		list_cfg[lists] = (struct cfg *) do_malloc(sizeof(struct cfg));

		if (sscanf(line, "%6s %48s %3s %50[^\n]\n", list_cfg[lists]->name, tmpadr, tmp_flags, list_cfg[lists]->desc) != 4) {
			print_error("%s: Missing fields in line %d\n", fname, lineno);
			return (-1);
		}

		if (sscanf(tmpadr, "%6[^@]@%40s", list_cfg[lists]->owner, list_cfg[lists]->owner_bbs) == 2)
			strupr(list_cfg[lists]->owner_bbs);
		else
			strcpy(list_cfg[lists]->owner_bbs, bbs);
		strupr(list_cfg[lists]->owner);

		list_cfg[lists]->access = 0;
		for (i = 0; tmp_flags[i]; i++)
			switch (toupper(tmp_flags[i])) {
			case 'R':
				list_cfg[lists]->access |= ACC_READ;
				break;
			case 'W':
				list_cfg[lists]->access |= ACC_RDWR;
				break;
			case 'P':
				list_cfg[lists]->access |= ACC_PRIV;
				break;
			case 'A':
				list_cfg[lists]->access |= ACC_ALL;
				break;
			default:
				print_error("%s: Invalid access type '%c' in line %d\n", fname, tmp_flags[i], lineno);
				return (-1);
			}

		if (list_cfg[lists]->access & ACC_READ && list_cfg[lists]->access & ACC_RDWR) {
			print_error("%s: Both access types `r' and `w' specified in line %d\n", fname, lineno);
			return (-1);
		}

		if (list_cfg[lists]->access == ACC_PRIV) {
			print_error("%s: Access type `r' or `w' missing in line %d\n", fname, lineno);
			return (-1);
		}

		lists++;
	}

	fclose(fp);
	return (0);
}

int check_list(char *name)
{
	int i;

	for (i = 0; i < lists; i++)
		if (!strcmp(name, list_cfg[i]->name))
			return (i);

	return (-1);
}

void print_acc(FILE * fp, int flags)
{
	if (flags & ACC_READ)
		fputs("R", fp);
	if (flags & ACC_RDWR)
		fputs("W", fp);
	if (flags & ACC_ALL)
		fputs("A", fp);
	if (flags & ACC_PRIV)
		fputs("/Priv", fp);
	else
		fputs("     ", fp);
	if (!(flags & ACC_ALL))
		fputs(" ", fp);
}

int num_users(char *listname)
{
	char fname[20];
	FILE *fp;
	int cnt = 0, ch;

	build_path(fname, listname, "USR");
	if ((fp = fopen(fname, "r")) == NULL)
		return (0);
	while ((ch = fgetc(fp)) != EOF) {
		if (ch == '\n')
			cnt++;
	}

	fclose(fp);
	return (cnt);
}

int add_user(char *listname, char *user, char *home)
{
	char buf[20];
	FILE *fp;

	build_path(buf, listname, "USR");
	if ((fp = fopen(buf, "a")) == NULL) {
		print_error("Error opening %s while adding a user: %s\n", buf, strerror(errno));
		return (-1);
	}
	if (home && *home)
		sscanf(home, "%6[^.].", buf);
	else
		strcpy(buf, bbs);
	fprintf(fp, "%s @ %s\n", user, buf);
	fclose(fp);

	return (0);
}

int remove_user(char *listname, char *user)
{
	char fname[20], arg1[7], line[80];
	FILE *fp, *tmp;

	build_path(fname, listname, "USR");
	if ((fp = fopen(fname, "r")) == NULL) {
		print_error("Error opening %s while removing a user: %s\n", fname, strerror(errno));
		return (-1);
	}

	if ((tmp = fopen("MLISTRMV.$$$", "w")) == NULL) {
		print_error("Error opening temp file while removing a user: %s\n", strerror(errno));
		return (-1);
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (sscanf(line, "%6s", arg1) == 1 && strcmp(user, arg1))
			if (fputs(line, tmp) == EOF) {
				print_error("Error writing to temp file while removing a user: %s\n", strerror(errno));
				return (-1);
			}
	}

	if (ferror(fp)) {
		print_error("Error reading from %s while removing a user: %s\n", fname, strerror(errno));
		return (-1);
	}

	fclose(fp);
	fclose(tmp);

	if (unlink(fname) || rename("MLISTRMV.$$$", fname)) {
		print_error("Error deleting %s, or renaming temp file to %s while removing a user: %s\n", fname, fname, strerror(errno));
		return (-1);
	}

	return (0);
}

int check_user(char *listname, char *user)
{
	char fname[20], arg1[7], line[80];
	FILE *fp;

	build_path(fname, listname, "USR");
	if ((fp = fopen(fname, "r")) == NULL) {
		print_error("Error opening %s: %s\n", fname, strerror(errno));
		return (-1);
	}

	while (fgets(line, sizeof(line), fp) != NULL) {
		if (sscanf(line, "%6s", arg1) != 1)
			continue;
		if (!strcmp(user, arg1)) {
			fclose(fp);
			return (1);
		}
	}

	fclose(fp);
	return (0);
}

void print_info(FILE * fp)
{
	int i;

	fputs("Active mailing lists:\n\n"
		"List   Owner  Access  Users Description\n"
		"---------------------------------------\n", fp);

	for (i = 0; i < lists; i++) {
		fprintf(fp, "%-6s %-6s ", list_cfg[i]->name, list_cfg[i]->owner);
		print_acc(fp, list_cfg[i]->access);
		fprintf(fp, " %5d %s\n", num_users(list_cfg[i]->name), list_cfg[i]->desc);
	}

	fprintf(fp, "\nR = Read Only - Only owner can post\n"
		"W = Read/Write - Posting OK\n"
		"A = Posting OK, even if you're not subscribed to the list\n"
		"Priv = Private list\n");
}

int main(int argc, char **argv)
{
	FILE *fp;
	char to[7];
	char sender[7];
	char home_bbs[41];
	char subject[80];
	long ofs;

#ifdef __MSDOS__
	if ((prgname = strrchr(argv[0], '.')) != NULL)
		*prgname = '\0';
#endif

	if ((prgname = strrchr(argv[0], PATH_SEP)) == NULL)
		prgname = argv[0];
	else
		prgname++;

	strupr(prgname);

	if (argc == 2 && !isatty(1)) {
		strcpy(subject, prgname);
		strcat(subject, ".ERR");
		close(2);
		open(subject, O_WRONLY | O_CREAT | O_APPEND, 0666);
	}

	if (read_cfg())
		return (1);

	if (!lists) {
		print_error("No lists defined!\n");
		return (1);
	}

	if (argc != 2) {
		printf(VERSTR, prgname);
		print_info(stdout);
		return (0);
	}

	if ((fp = fopen(argv[1], "r")) == NULL) {
		print_error("Error opening server input file %s: %s\n", argv[1], strerror(errno));
		return (1);
	}

	if ((ofs = read_input(fp, to, sender, home_bbs, subject)) == -1) {
		print_error("Invalid server input file\n");
		return (1);
	}

	if ((mail_in = fopen(mailin_file, "a")) == NULL) {
		print_error("Error opening %s: %s\n", mailin_file, strerror(errno));
		exit (1);
	}

	if (!strcmp(to, prgname)) {
		fprintf(mail_in, "SP %s @ %s < %s\n"
			"%s results\n", sender, home_bbs, bbs, prgname);
		if (handle_commands(fp, sender, home_bbs))
			return (1);
		fprintf(mail_in, "----\nFor help, send a message to %s @ %s\n"
			"with the command `HELP' in the body of the message.\n/EX\n", prgname, bbs);
	}
	else {
		if (copy_message(fp, ofs, to, sender, home_bbs, subject))
			return (1);
	}

#ifndef __MSDOS__
	sprintf(subject, "%s.%d", prgname, getpid());
#else
	strcpy(subject, "MLISTMSG.$$$");
#endif
	if (copy_file(subject, mail_in) == 0) {
		fclose(mail_in);
		unlink(subject);
	}

#ifndef TEST_VERSION
	unlink(argv[1]);
#endif

	return (0);
}
